#include <cstring>
#include "barnes_hut.h"

BarnesHutTree::
BarnesHutTree(void)
   : xmin(0,0), xmax(0,0), total_mass(0), centre_of_mass(0,0), bounding_radius(0), leaf(true)
{
   children[0]=children[1]=0;
}

BarnesHutTree::
~BarnesHutTree(void)
{
   if(children[0]) delete children[0];
   if(children[1]) delete children[1];
}

void BarnesHutTree::
clear(void)
{
   if(!leaf){
      if(children[0]){ delete children[0]; children[0]=0; }
      if(children[1]){ delete children[1]; children[1]=0; }
   }
   xmin=Vec2d(0,0);
   xmax=Vec2d(0,0);
   total_mass=0;
   centre_of_mass=Vec2d(0,0);
   bounding_radius=0;
   leaf=true;
}

void BarnesHutTree::
construct_from_points(int n, const Vec2d *x, const double *mass, int max_depth, int max_points_per_leaf)
{
   assert(n>=1 && x && mass && max_depth>=0);
   clear();
   // set up a list of all point indices (initially just in order 0, 1, ..., n-1)
   int *indices=new int[n];
   for(int i=0; i<n; ++i) indices[i]=i;
   construct_recursively(n, x, mass, max_depth, max_points_per_leaf, 0, n, indices);
   delete[] indices;
}

void BarnesHutTree::
construct_recursively(int n, const Vec2d *x, const double *mass,
                      int max_depth, int max_points_per_leaf, int current_depth, int num_indices, int *indices)
{
   assert(max_points_per_leaf>0);
   assert(current_depth<=max_depth);
   // find bounding box for all points
   xmin=x[indices[0]];
   xmax=x[indices[0]];
   for(int i=1; i<num_indices; ++i){
      Vec2d xi=x[indices[i]];
      if(xi[0]<xmin[0])      xmin[0]=xi[0];
      else if(xi[0]>xmax[0]) xmax[0]=xi[0];
      if(xi[1]<xmin[1])      xmin[1]=xi[1];
      else if(xi[1]>xmax[1]) xmax[1]=xi[1];
   }
   // base case of recursion (keep *this as a leaf, with the given points)
   if(current_depth==max_depth || num_indices<=max_points_per_leaf || xmax==xmin){
      assert(leaf && total_mass==0 && centre_of_mass==Vec2d(0,0));
      assert(!children[0] && !children[1]);
      point_indices.resize(num_indices);
      for(int i=0; i<num_indices; ++i){
         int j=indices[i];
         point_indices[i]=j;
         total_mass+=mass[j];
         centre_of_mass+=mass[j]*x[j];
      }
      assert(total_mass>0);
      centre_of_mass/=total_mass;
      // find tight bound on radius of this leaf
      double r2=0;
      for(int i=0; i<num_indices; ++i){
         int j=indices[i];
         double d2=dist2(centre_of_mass, x[j]);
         if(d2>r2) r2=d2;
      }
      bounding_radius=std::sqrt(r2);

   }else{ // we'll split up the indices we've been given to child nodes
      leaf=false;
      // figure out which axis to split along
      int axis=0;
      if(xmax[0]-xmin[0]<xmax[1]-xmin[1]) axis=1;
      // pick a splitting value: note, since axis bounds were tight, there must be at least one point on either side of split 
      double split=0.5*(xmax[axis]+xmin[axis]);
      // do the quicksort-style in-place split of the indices (could be improved)
      int i=0, j=num_indices-1;
      for(;;){
         assert(i<num_indices);
         while(x[indices[i]][axis]<=split){
            ++i;
            assert(i<num_indices);
         }
         assert(j>=0);
         while(x[indices[j]][axis]>split){
            --j;
            assert(j>=0);
         }
         if(i<j){
            std::swap(indices[i], indices[j]);
            ++i;
            --j;
         }else{
            break;
         }
      }
      assert(i==j+1);
      assert(i>0 && i<num_indices);
      // now indices[0] to indices[i-1] inclusive should all be located <=split (i points)
      // and indices[i] to indices[num_indices-1] inclusive should be located >split (num_indices-i points) 
      children[0]=new BarnesHutTree;
      children[0]->construct_recursively(n, x, mass, max_depth, max_points_per_leaf, current_depth+1, i, indices);
      children[1]=new BarnesHutTree;
      children[1]->construct_recursively(n, x, mass, max_depth, max_points_per_leaf, current_depth+1, num_indices-i, indices+i);
      // and get the mass information from the children
      total_mass=children[0]->total_mass+children[1]->total_mass;
      centre_of_mass=(children[0]->total_mass*children[0]->centre_of_mass
                      + children[1]->total_mass*children[1]->centre_of_mass)/total_mass;
      // get a cheap upper bound on bounding_radius from bounding box
      double d0=std::max( xmax[0]-centre_of_mass[0], centre_of_mass[0]-xmin[0] ),
             d1=std::max( xmax[1]-centre_of_mass[1], centre_of_mass[1]-xmin[1] );
      double boxbound=std::sqrt(d0*d0+d1*d1);
      // get another bound from children
      double childbound=std::max(dist(centre_of_mass, children[0]->centre_of_mass) + children[0]->bounding_radius,
                                 dist(centre_of_mass, children[1]->centre_of_mass) + children[1]->bounding_radius);
      // take the tighter bound
      bounding_radius=std::min(boxbound, childbound);
   }
}

void BarnesHutTree::
evaluate_gravity(int n, const Vec2d *x, const double *mass, Vec2d *accel, const double error_tolerance,
                 const double gravitational_constant)
{
   assert(accel);
   for(int i=0; i<n; ++i){
      accel[i]=Vec2d(0,0);
      add_gravity(n, x, mass, accel, i, error_tolerance); // recursive call on tree
      accel[i]*=gravitational_constant;
   }
}

void BarnesHutTree::
add_gravity(int n, const Vec2d *x, const double *mass, Vec2d *accel, int i, const double error_tolerance)
{
   // You will need to fill in the code for this function.
   // It should (recursively) add this node of the tree's contribution to the acceleration of point i
   // The parameter error_tolerance should guide when to use the Barnes-Hut approximation or when to
   // recurse further.
}

