#include <cmath>
#include <iostream>
#include <ctime>
#include "brute_force.h"
#include "barnes_hut.h"
#include "writesgi.h"

// return a randomly generated vector uniformly distributed within the unit disc
Vec2d random_disc_vector(void)
{
   Vec2d r;
   do{ // rejection sampling
      r[0]=2*(rand()/(double)RAND_MAX)-1;
      r[1]=2*(rand()/(double)RAND_MAX)-1;
   }while(mag2(r)>1);
   return r;
}

// if you want you can use these to dump out pictures (in SGI format) of your simulation
void draw_points(int n, const Vec2d *x, double R, int step);
void draw_tree(const BarnesHutTree &tree, int n, const Vec2d *x, int step);

int main()
{
   int n=1000;
   double *mass=new double[n];
   Vec2d *x=new Vec2d[n];
   Vec2d *v=new Vec2d[n];
   Vec2d *a=new Vec2d[n];
   double t=0;

   // set up initial conditions
   const double G=6.67428e-11; // gravitational constant
   const double R=5e12; // approximate radius of solar system (m)
   const double M=1e30; // approximate mass of solar system (kg)
   // the central mass point will have most of the total mass (like our Sun)
   mass[0]=0.99*M;
   x[0]=Vec2d(0,0);
   v[0]=Vec2d(0,0);
   // evenly distribute the rest of the mass to the other points
   for(int p=1; p<n; ++p){
      mass[p]=0.01*M/n;
      x[p]=R*random_disc_vector(); // somewhere chosen uniformly from rough bound on solar system
      x[p][1]*=0.25; // but squish it into an ellipse
      v[p]=std::sqrt(G*M/mag(x[p]))*perp(x[p])/mag(x[p]);
   }

   //draw_points(n, x, R, 0);

   // loop
   double dt=60*60*24*30; // one month in seconds
   for(int step=1; step<1000; ++step){
      std::cout<<step<<": ";
      clock_t t0=std::clock();
      /*
      BarnesHutTree tree;
      tree.construct_from_points(n, x, mass, 100, 1);
      tree.evaluate_gravity(n, x, mass, a, 0.05);
      */
      evaluate_gravity(n, mass, x, a);
      // Verlet integration for each particle
      for(int p=0; p<n; ++p){
         v[p]+=dt*a[p];
         x[p]+=dt*v[p];
      }
      t+=dt;
      clock_t t1=std::clock();
      std::cout<<(t1-t0)/(double)CLOCKS_PER_SEC<<" seconds"<<std::endl;
      //draw_points(n, x, R, step);
      //draw_tree(tree, n, x, step);
   }

   delete[] mass;
   delete[] x;
   delete[] v;
   delete[] a;
   return 0;
}

void draw_points(int n, const Vec2d *x, double R, int step)
{
   static Array2<Vec3f> image(512,512);
   image.assign(Vec3f(0,0,0));
   for(int p=0; p<n; ++p){
      int i=(int)(image.nx*(x[p][0]/(4*R)+0.5));
      int j=(int)(image.ny*(x[p][1]/(4*R)+0.5));
      if(i>=0 && i<image.nx && j>=0 && j<image.ny){
         image(i,j)=Vec3f(1,1,1);
      }
   }
   write_sgi(image, "gravity%04d.sgi", step);
}

void draw_tree_recursively(const BarnesHutTree &tree, Array2<Vec3f> &image, int res, double x0, double y0, double scale)
{
   int i0=(int)(scale*(tree.xmin[0]-x0)); if(i0<0) i0=0; else if(i0>=res) i0=res-1;
   int j0=(int)(scale*(tree.xmin[1]-y0)); if(j0<0) j0=0; else if(j0>=res) j0=res-1;
   int i1=(int)(scale*(tree.xmax[0]-x0)); if(i1<0) i1=0; else if(i1>=res) i1=res-1;
   int j1=(int)(scale*(tree.xmax[1]-y0)); if(j1<0) j1=0; else if(j1>=res) j1=res-1;
   for(int s=i0; s<=i1; ++s){
      image(s,j0)=Vec3f(0,0,0);
      image(s,j1)=Vec3f(0,0,0);
   }
   for(int s=j0; s<=j1; ++s){
      image(i0,s)=Vec3f(0,0,0);
      image(i1,s)=Vec3f(0,0,0);
   }
   if(!tree.leaf){
      draw_tree_recursively(*tree.children[0], image, res, x0, y0, scale);
      draw_tree_recursively(*tree.children[1], image, res, x0, y0, scale);
   }
}

void draw_tree(const BarnesHutTree &tree, int n, const Vec2d *x, int step)
{
   int res=1000;
   static Array2<Vec3f> image(res,res);
   image.assign(Vec3f(1,1,1));
   double x0=tree.xmin[0], y0=tree.xmin[1];
   double scale=res/std::max(tree.xmax[0]-x0, tree.xmax[1]-y0);
   draw_tree_recursively(tree, image, res, x0, y0, scale);
   for(int p=0; p<n; ++p){
      int i=scale*(x[p][0]-x0); if(i<0) i=0; else if(i>=res) i=res-1;
      int j=scale*(x[p][1]-y0); if(j<0) j=0; else if(j>=res) j=res-1;
      image(i,j)=Vec3f(1,0,0);
   }
   write_sgi(image, "tree%04d.sgi", step);
}

